#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

import importlib.util
from inspect import isclass
from types import GenericAlias
from typing import (
    Any,
    Callable,
    Mapping,
    Optional,
    Protocol,
    Sequence,
    Union,
    get_args,
    get_origin,
)

from meatie import AsyncResponse, Response
from meatie.internal.types import T

from . import (
    BytesAdapter,
    ClientResponseAdapter,
    JsonAdapter,
    NoneAdapter,
    StringAdapter,
    TypeAdapter,
)


def _is_model_type_no_pydantic(value: type[Any]) -> bool:  # pragma: no cover
    return False


class PydanticTypeAdapterFactory(Protocol):
    def __call__(self, model_cls: type[T]) -> TypeAdapter[T]:
        ...

    @staticmethod
    def is_model_type(model_cls: type[Any]) -> bool:
        ...


def _resolve_pydantic_type_adapter_factory() -> (
    Optional[PydanticTypeAdapterFactory]
):  # pragma: no cover
    module_spec = importlib.util.find_spec("pydantic")

    if module_spec is None:
        return None

    try:
        from .pydantic_v2 import PydanticV2TypeAdapterFactory

        return PydanticV2TypeAdapterFactory()
    except ImportError:
        from .pydantic_v1 import PydanticV1TypeAdapterFactory

        return PydanticV1TypeAdapterFactory()


_PydanticTypeAdapterFactory: Optional[
    PydanticTypeAdapterFactory
] = _resolve_pydantic_type_adapter_factory()


def _resolve_is_model_type() -> Callable[[type[Any]], bool]:
    if _PydanticTypeAdapterFactory is None:
        return _is_model_type_no_pydantic
    return _PydanticTypeAdapterFactory.is_model_type


_is_model_type = _resolve_is_model_type()


def get_adapter(value_type: Union[type[T], GenericAlias, None]) -> TypeAdapter[T]:
    if value_type is None:
        return NoneAdapter  # type: ignore[return-value]

    if value_type is bytes:
        return BytesAdapter  # type: ignore[return-value]

    if value_type is str:
        return StringAdapter  # type: ignore[return-value]

    if isinstance(value_type, GenericAlias):
        if _PydanticTypeAdapterFactory is None:
            return JsonAdapter

        origin = get_origin(value_type)
        args = get_args(value_type)
        if issubclass(origin, Sequence):
            if _is_model_type(args[0]):
                return _PydanticTypeAdapterFactory(value_type)  # type: ignore[arg-type]
            return JsonAdapter

        if issubclass(origin, Mapping):
            if _is_model_type(args[1]):
                return _PydanticTypeAdapterFactory(value_type)  # type: ignore[arg-type]
            return JsonAdapter

        return JsonAdapter

    if _is_model_type(value_type):
        if _PydanticTypeAdapterFactory is None:
            return JsonAdapter
        return _PydanticTypeAdapterFactory(value_type)

    if isclass(value_type) and (value_type is Response or value_type is AsyncResponse):
        return ClientResponseAdapter  # type: ignore[return-value]

    return JsonAdapter



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.


import abc
from typing import Callable

from .context import RetryContext

Condition = Callable[[RetryContext], bool]


class BaseCondition:
    @abc.abstractmethod
    def __call__(self, ctx: RetryContext) -> bool:
        pass

    def __and__(self, other: Condition) -> Condition:
        return AndCondition(self, other)

    def __or__(self, other: Condition) -> Condition:
        return OrCondition(self, other)


class AndCondition(BaseCondition):
    def __init__(self, left: Condition, right: Condition) -> None:
        self._left = left
        self._right = right

    def __call__(self, ctx: RetryContext) -> bool:
        return self._left(ctx) and self._right(ctx)


class OrCondition(BaseCondition):
    def __init__(self, left: Condition, right: Condition) -> None:
        self._left = left
        self._right = right

    def __call__(self, ctx: RetryContext) -> bool:
        return self._left(ctx) or self._right(ctx)



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

# isort:skip_file
import importlib.metadata
from .types import Request, Response, AsyncResponse, Method, Duration, Time, INF, MINUTE, HOUR, DAY
from .api_reference import api_ref
from .error import (
    MeatieError,
    RetryError,
    RequestError,
    RateLimitExceeded,
    TransportError,
    ProxyError,
    ServerError,
    Timeout,
    HttpStatusError,
    ResponseError,
    ParseResponseError,
)
from .internal.retry import (
    RetryContext,
    BaseCondition,
    Condition,
    after,
    after_attempt,
    always,
    never,
    has_status,
    has_exception_type,
    has_exception_cause_type,
    zero,
    uniform,
    exponential,
    fixed,
    jit,
)
from .internal.cache import Cache
from .internal.limit import Limiter, Rate
from .client import BaseClient
from .descriptor import EndpointDescriptor, Context
from .aio import BaseAsyncClient, AsyncEndpointDescriptor, AsyncContext
from .option import (
    limit,
    cache,
    retry,
    private,
    body,
)
from .endpoint import endpoint

__all__ = [
    "Duration",
    "Time",
    "INF",
    "MINUTE",
    "HOUR",
    "DAY",
    "Method",
    "Request",
    "Response",
    "AsyncResponse",
    "api_ref",
    "MeatieError",
    "RetryError",
    "RequestError",
    "RateLimitExceeded",
    "TransportError",
    "ProxyError",
    "ServerError",
    "Timeout",
    "HttpStatusError",
    "ResponseError",
    "ParseResponseError",
    "RetryContext",
    "BaseCondition",
    "Condition",
    "zero",
    "uniform",
    "exponential",
    "fixed",
    "jit",
    "never",
    "after",
    "after_attempt",
    "always",
    "has_status",
    "has_exception_type",
    "has_exception_cause_type",
    "Cache",
    "Limiter",
    "Rate",
    "BaseClient",
    "Context",
    "EndpointDescriptor",
    "BaseAsyncClient",
    "AsyncEndpointDescriptor",
    "AsyncContext",
    "retry",
    "limit",
    "cache",
    "private",
    "body",
    "endpoint",
]

__version__ = importlib.metadata.version("meatie")



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

from meatie import AsyncResponse, Response


class _NoneAdapter:
    @staticmethod
    def from_response(_: Response) -> None:  # pragma: no cover
        return None

    @staticmethod
    async def from_async_response(_: AsyncResponse) -> None:  # pragma: no cover
        return None

    @staticmethod
    def to_content(_: None) -> None:  # pragma: no cover
        return None


NoneAdapter = _NoneAdapter()



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.


from meatie import AsyncResponse, Response

from .types import TypeAdapter


class _BytesAdapter:
    @staticmethod
    def from_response(response: Response) -> bytes:
        return response.read()

    @staticmethod
    async def from_async_response(response: AsyncResponse) -> bytes:
        return await response.read()

    @staticmethod
    def to_content(value: bytes) -> bytes:
        return value


BytesAdapter: TypeAdapter[bytes] = _BytesAdapter()



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

import inspect
from types import GenericAlias
from typing import (
    Any,
    Generic,
    Iterable,
    Optional,
)

from meatie import Method, Request
from meatie.api_reference import ApiReference
from meatie.internal.adapter import JsonAdapter, TypeAdapter, get_adapter
from meatie.internal.types import PT, RequestBodyType, T
from typing_extensions import Callable, Self, Union, get_type_hints

from . import Kind, Parameter, PathTemplate


class RequestTemplate(Generic[RequestBodyType]):
    __slots__ = (
        "method",
        "template",
        "params",
        "request_encoder",
        "__param_by_name",
    )

    def __init__(
        self,
        template: PathTemplate,
        params: list[Parameter],
        request_encoder: TypeAdapter[RequestBodyType],
        method: Optional[Method],
    ) -> None:
        self.method = method
        self.template = template
        self.params = params
        self.request_encoder = request_encoder

        self.__param_by_name: dict[str, Parameter] = {}
        for param in self.params:
            self.__param_by_name[param.name] = param

    def build_request(self, *args: Any, **kwargs: Any) -> Request:
        if self.method is None:
            raise RuntimeError("'method' is None")

        if len(args) > len(self.params):
            raise RuntimeError("Too many arguments passed to the function.")

        value_by_param = {}

        # set arguments passed as args to corresponding values
        for index, value in enumerate(args):
            param = self.params[index]
            value_by_param[param] = value

        # set arguments to default parameter values if arguments were not passed already as args
        for param in self.params:
            if param.default_value is None or param in value_by_param:
                continue
            value_by_param[param] = param.default_value

        # override default values by actual values passed in kwargs
        for name, value in kwargs.items():
            param_opt = self.__param_by_name.get(name)
            if param_opt is None:
                raise ValueError(f"Parameter '{name}' is not mentioned in the endpoint definition.")
            value_by_param[param_opt] = value

        path_kwargs = {}
        query_kwargs = {}
        body_value: Any = None
        for param, value in value_by_param.items():
            if param.kind == Kind.PATH:
                if param.formatter is not None:
                    value = param.formatter(value)
                path_kwargs[param.api_ref] = value
                continue

            if param.kind == Kind.QUERY:
                # emit query parameters only if underlying value is not None
                if value is not None:
                    if param.formatter is not None:
                        value = param.formatter(value)
                    query_kwargs[param.api_ref] = value
                continue

            if param.kind == Kind.BODY:
                if param.formatter is not None:
                    body_value = param.formatter(value)
                else:
                    body_value = value
                continue

            raise NotImplementedError(f"Kind {param.kind} is not supported")  # pragma: no cover

        path = self.template.format(**path_kwargs)

        if body_value is not None:
            body_json = self.request_encoder.to_content(body_value)
        else:
            body_json = None

        return Request(
            method=self.method, path=path, params=query_kwargs, headers={}, json=body_json
        )

    @classmethod
    def from_callable(
        cls, func: Callable[PT, T], template: PathTemplate, method: Optional[Method]
    ) -> Self:
        signature = inspect.signature(func)
        type_hints = get_type_hints(func)
        return cls.from_signature(signature, type_hints, template, method)

    @classmethod
    def from_signature(
        cls,
        signature: inspect.Signature,
        type_hints: dict[str, Union[type, GenericAlias, None]],
        template: PathTemplate,
        method: Optional[Method],
    ) -> Self:
        parameters = []
        request_encoder: TypeAdapter[Any] = JsonAdapter
        for param_name in type_hints:
            if param_name == "return":
                continue

            param_type = type_hints[param_name]
            sig_param = signature.parameters[param_name]
            api_ref = ApiReference.from_signature(sig_param)
            assert api_ref.name is not None

            kind = Kind.QUERY
            if api_ref.name == "body":
                kind = Kind.BODY
                request_encoder = get_adapter(param_type)
            elif api_ref.name in template:
                kind = Kind.PATH

            default_value = None
            if sig_param.default is not inspect.Parameter.empty:
                default_value = sig_param.default

            formatter = None
            if api_ref.fmt is not None:
                formatter = api_ref.fmt
            parameter = Parameter(kind, param_name, api_ref.name, default_value, formatter)
            parameters.append(parameter)

        return cls.validate_object(template, parameters, signature, request_encoder, method)

    @classmethod
    def validate_object(
        cls,
        template: PathTemplate,
        params: Iterable[Parameter],
        signature: inspect.Signature,
        request_encoder: TypeAdapter[RequestBodyType],
        method: Optional[Method],
    ) -> Self:
        template_str = str(template)
        if template_str == "":
            raise ValueError("'path' is empty")

        visited_api_refs = set()
        for param in params:
            if param.name == "self":
                raise ValueError("Parameter 'self' is forbidden")

            if param.api_ref in visited_api_refs:
                raise ValueError(f"Multiple parameters have name '{param.api_ref}'")

            if param.kind == Kind.PATH and param.api_ref not in template:
                raise ValueError(
                    f"Parameter '{param.api_ref}' is not present in the path '{template}'"
                )

            visited_api_refs.add(param.api_ref)

        missing_api_refs = set(template.parameters) - visited_api_refs
        for api_ref in missing_api_refs:
            raise ValueError(
                f"Parameter '{api_ref}' is not present in the method signature '{signature}'"
            )

        return cls(template, list(params), request_encoder, method)



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.
from typing import (
    Any,
    Awaitable,
    Callable,
    Generic,
    Iterable,
    Optional,
    Union,
    overload,
)

from typing_extensions import Self

from meatie import BaseClient, Request, Response
from meatie.internal.adapter import TypeAdapter
from meatie.internal.template import RequestTemplate, get_method
from meatie.internal.types import PT, ResponseBodyType

Operator = Callable[["Context[ResponseBodyType]"], ResponseBodyType]


class EndpointDescriptor(Generic[PT, ResponseBodyType]):
    def __init__(
        self, template: RequestTemplate[Any], response_decoder: TypeAdapter[ResponseBodyType]
    ) -> None:
        self.template = template
        self.response_decoder = response_decoder
        self.get_json: Optional[Callable[[Any], dict[str, Any]]] = None
        self.get_text: Optional[Callable[[Any], str]] = None
        self.__operator_by_priority: dict[int, Operator[ResponseBodyType]] = {}

    def __set_name__(self, owner: type[object], name: str) -> None:
        if self.template.method is not None:
            return

        self.template.method = get_method(name)

    def register_operator(self, priority: int, operator: Operator[ResponseBodyType]) -> None:
        self.__operator_by_priority[priority] = operator

    @overload
    def __get__(self, instance: None, owner: None) -> Self:
        ...

    @overload
    def __get__(
        self, instance: BaseClient, owner: type[object]
    ) -> Callable[PT, Awaitable[ResponseBodyType]]:
        ...

    def __get__(
        self, instance: Optional[BaseClient], owner: Optional[type[object]] = None
    ) -> Union[Self, Callable[PT, Awaitable[ResponseBodyType]]]:
        if instance is None or owner is None:
            return self

        priority_operator_pair = list(self.__operator_by_priority.items())
        priority_operator_pair.sort()
        operators = [operator for _, operator in priority_operator_pair]

        return BoundEndpointDescriptor(  # type: ignore[return-value]
            instance, operators, self.template, self.response_decoder, self.get_json, self.get_text
        )


class Context(Generic[ResponseBodyType]):
    def __init__(
        self,
        client: BaseClient,
        operators: list[Operator[ResponseBodyType]],
        request: Request,
    ) -> None:
        self.__client = client
        self.__operators = operators
        self.__next_step = 0

        self.__request = request
        self.response: Optional[Response] = None

    @property
    def client(self) -> BaseClient:
        return self.__client

    @property
    def request(self) -> Request:
        return self.__request

    def proceed(self) -> ResponseBodyType:
        if self.__next_step >= len(self.__operators):
            raise RuntimeError("No more step to process request")

        current_step = self.__next_step
        self.__next_step += 1
        current_operator = self.__operators[current_step]
        try:
            result = current_operator(self)
            return result
        finally:
            self.__next_step = current_step


class BoundEndpointDescriptor(Generic[PT, ResponseBodyType]):
    def __init__(
        self,
        instance: BaseClient,
        operators: Iterable[Operator[ResponseBodyType]],
        template: RequestTemplate[Any],
        response_decoder: TypeAdapter[ResponseBodyType],
        get_json: Optional[Callable[[Any], dict[str, Any]]],
        get_text: Optional[Callable[[Any], str]],
    ) -> None:
        self.__instance = instance
        self.__operators = list(operators)
        self.__operators.append(self.__send_request)
        self.__template = template
        self.__response_decoder = response_decoder
        self.__get_json = get_json
        self.__get_text = get_text

    def __call__(self, *args: PT.args, **kwargs: PT.kwargs) -> ResponseBodyType:
        request = self.__template.build_request(*args, **kwargs)
        context = Context[ResponseBodyType](self.__instance, self.__operators, request)
        return context.proceed()

    def __send_request(self, context: Context[ResponseBodyType]) -> ResponseBodyType:
        response = self.__instance.send(context.request)
        if self.__get_json is not None:
            response.get_json = self.__get_json
        if self.__get_text is not None:
            response.get_text = self.__get_text
        context.response = response
        return self.__response_decoder.from_response(response)



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.
import asyncio
import time
from typing import Awaitable, Callable, Union

from meatie import Duration
from meatie.aio import AsyncContext, AsyncEndpointDescriptor
from meatie.descriptor import Context, EndpointDescriptor
from meatie.internal.limit import Tokens
from meatie.internal.types import PT, T

__all__ = ["limit"]


class LimitOption:
    __PRIORITY = 80

    def __init__(
        self,
        tokens: Tokens,
        sleep_func: Union[Callable[[float], Union[None, Awaitable[None]]], None] = None,
    ) -> None:
        self.tokens = tokens
        self.sleep_func = sleep_func

    def __call__(
        self, descriptor: Union[EndpointDescriptor[PT, T], AsyncEndpointDescriptor[PT, T]]
    ) -> None:
        if isinstance(descriptor, EndpointDescriptor):
            return self.__sync_descriptor(descriptor)
        return self.__async_descriptor(descriptor)

    def __sync_descriptor(self, descriptor: EndpointDescriptor[PT, T]) -> None:
        if self.tokens <= 0.0:
            return

        sleep_func: Callable[[float], None] = time.sleep
        if self.sleep_func is not None:
            sleep_func = self.sleep_func  # type: ignore[assignment]
        operator = LimitOperator(self.tokens, sleep_func)
        descriptor.register_operator(LimitOption.__PRIORITY, operator)

    def __async_descriptor(self, descriptor: AsyncEndpointDescriptor[PT, T]) -> None:
        if self.tokens <= 0.0:
            return

        sleep_func: Callable[[float], Awaitable[None]] = asyncio.sleep
        if self.sleep_func is not None:
            sleep_func = self.sleep_func  # type: ignore[assignment]
        operator = AsyncLimitOperator(self.tokens, sleep_func)
        descriptor.register_operator(LimitOption.__PRIORITY, operator)


limit = LimitOption


class LimitOperator:
    def __init__(self, tokens: Tokens, sleep_func: Callable[[Duration], None]) -> None:
        self.tokens = tokens
        self.sleep_func = sleep_func

    def __call__(self, ctx: Context[T]) -> T:
        current_time = time.monotonic()
        reservation = ctx.client.limiter.reserve_at(current_time, self.tokens)
        delay = reservation.ready_at - current_time
        if delay > 0:
            self.sleep_func(delay)

        return ctx.proceed()


class AsyncLimitOperator:
    def __init__(self, tokens: Tokens, sleep_func: Callable[[Duration], Awaitable[None]]) -> None:
        self.tokens = tokens
        self.sleep_func = sleep_func

    async def __call__(self, ctx: AsyncContext[T]) -> T:
        current_time = time.monotonic()
        reservation = ctx.client.limiter.reserve_at(current_time, self.tokens)
        delay = reservation.ready_at - current_time
        if delay > 0:
            await self.sleep_func(delay)

        return await ctx.proceed()



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

# isort:skip_file

from .client import BaseAsyncClient
from .descriptor import AsyncEndpointDescriptor, AsyncContext

__all__ = [
    "BaseAsyncClient",
    "AsyncEndpointDescriptor",
    "AsyncContext",
]



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.


from typing import Union

from ... import ResponseError
from . import BaseCondition, RetryContext

__all__ = ["always", "has_status", "has_exception_type", "has_exception_cause_type"]


class RetryAlways(BaseCondition):
    def __call__(self, ctx: RetryContext) -> bool:
        return True


class RetryOnStatus(BaseCondition):
    def __init__(self, status: int) -> None:
        self.status = status

    def __call__(self, ctx: RetryContext) -> bool:
        if ctx.response is not None:
            return ctx.response.status == self.status

        if isinstance(ctx.error, ResponseError):
            return ctx.error.response.status == self.status

        return False


class RetryOnExceptionType(BaseCondition):
    def __init__(
        self, exc_types: Union[type[BaseException], tuple[type[BaseException], ...]]
    ) -> None:
        self.exc_types = exc_types

    def __call__(self, ctx: RetryContext) -> bool:
        return isinstance(ctx.error, self.exc_types)


class RetryOnExceptionCauseType(BaseCondition):
    def __init__(
        self, exc_types: Union[type[BaseException], tuple[type[BaseException], ...]]
    ) -> None:
        self.exc_types = exc_types

    def __call__(self, ctx: RetryContext) -> bool:
        exc = ctx.error
        while exc is not None:
            if isinstance(exc, self.exc_types):
                return True
            exc = exc.__cause__
        return False


always = RetryAlways()
has_status = RetryOnStatus
has_exception_type = RetryOnExceptionType
has_exception_cause_type = RetryOnExceptionCauseType



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.
from enum import Enum
from typing import Any, Callable, Optional


class Kind(Enum):
    UNKNOWN = 0
    PATH = 1
    QUERY = 2
    BODY = 3


class Parameter:
    __slots__ = ("kind", "name", "api_ref", "default_value", "formatter")

    def __init__(
        self,
        kind: Kind,
        name: str,
        api_ref: str,
        default_value: Any = None,
        formatter: Optional[Callable[[Any], Any]] = None,
    ) -> None:
        self.kind = kind
        self.name = name
        self.api_ref = api_ref
        self.default_value = default_value
        self.formatter = formatter

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Parameter):
            return (
                self.name == other.name
                and self.kind == other.kind
                and self.api_ref == other.api_ref
                and self.default_value == other.default_value
            )
        return False



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

# isort:skip_file
from .types import TypeAdapter
from .bytes_ import BytesAdapter
from .json_ import JsonAdapter
from .client_response import ClientResponseAdapter
from .none_ import NoneAdapter
from .string_ import StringAdapter
from .factory import get_adapter

__all__ = [
    "TypeAdapter",
    "JsonAdapter",
    "NoneAdapter",
    "BytesAdapter",
    "StringAdapter",
    "ClientResponseAdapter",
    "get_adapter",
]



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.


from meatie import INF, Duration

from . import Tokens


class Rate:
    max: "Rate"
    __slots__ = ("__tokens_per_sec",)

    def __init__(self, tokens_per_sec: float) -> None:
        if tokens_per_sec <= 0:
            raise ValueError("'tokens_per_sec' must be positive")

        self.__tokens_per_sec = tokens_per_sec

    def duration_from_tokens(self, tokens: Tokens) -> Duration:
        return tokens / self.__tokens_per_sec

    def tokens_from_duration(self, duration: Duration) -> Tokens:
        return self.__tokens_per_sec * duration


Rate.max = Rate(INF)



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.
from typing import Any, Optional

import httpx
from meatie import BaseClient, Cache
from meatie.error import (
    MeatieError,
    ProxyError,
    RequestError,
    ServerError,
    Timeout,
    TransportError,
)
from meatie.types import Request
from typing_extensions import Self

from . import Response


class Client(BaseClient):
    def __init__(
        self,
        client: httpx.Client,
        client_params: Optional[dict[str, Any]] = None,
        local_cache: Optional[Cache] = None,
        limiter: Optional[Any] = None,
    ) -> None:
        super().__init__(local_cache, limiter)

        self.client = client
        self.client_params = client_params if client_params else {}

    def send(self, request: Request) -> Response:
        kwargs: dict[str, Any] = self.client_params.copy()

        if request.data is not None:
            kwargs["content"] = request.data

        if request.json is not None:
            kwargs["json"] = request.json

        if request.headers:
            kwargs["headers"] = request.headers

        if request.params:
            kwargs["params"] = request.params

        try:
            response = self.client.request(request.method, request.path, **kwargs)
        except (httpx.InvalidURL, httpx.UnsupportedProtocol) as exc:
            raise RequestError(exc) from exc
        except httpx.ProxyError as exc:
            raise ProxyError(exc) from exc
        except httpx.TimeoutException as exc:
            raise Timeout(exc) from exc
        except (httpx.NetworkError, httpx.RemoteProtocolError) as exc:
            raise ServerError(exc) from exc
        except (httpx.TooManyRedirects, httpx.ProtocolError) as exc:
            raise TransportError(exc) from exc
        except httpx.HTTPError as exc:
            raise MeatieError(exc) from exc
        return Response(response)

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Any,
    ) -> None:
        self.client.close()

    def close(self) -> None:
        self.client.close()



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

# isort: skip_file
from .method import get_method
from .parameter import Kind, Parameter
from .path import PathTemplate
from .request import RequestTemplate

__all__ = ["Kind", "Parameter", "PathTemplate", "RequestTemplate", "get_method"]



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

# isort:skip_file
from .response import Response
from .client import Client

__all__ = ["Response", "Client"]



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

# isort: skip_file

from .reservation import Reservation, Tokens
from .rate import Rate
from .limiter import Limiter

__all__ = ["Rate", "Tokens", "Reservation", "Limiter"]



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.


from typing import Optional, Union

from meatie import AsyncResponse, Response, Time


class RetryContext:
    __slots__ = ("attempt_number", "started_at", "error", "response")

    def __init__(
        self,
        attempt_number: int,
        started_at: Time,
        error: Optional[BaseException] = None,
        response: Optional[Union[Response, AsyncResponse]] = None,
    ) -> None:
        self.attempt_number = attempt_number
        self.started_at = started_at
        self.error: Optional[BaseException] = error
        self.response: Optional[Union[Response, AsyncResponse]] = response



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

import time

from . import BaseCondition, Condition, RetryContext

__all__ = ["after", "after_attempt", "never"]

from meatie import Duration


class StopAfterAttempt(BaseCondition):
    def __init__(self, attempts: int) -> None:
        self.attempts = attempts

    def __call__(self, ctx: RetryContext) -> bool:
        return ctx.attempt_number > self.attempts


class StopAfter(BaseCondition):
    def __init__(self, max_delay: Duration) -> None:
        self.max_delay = max_delay

    def __call__(self, ctx: RetryContext) -> bool:
        elapsed_time = time.monotonic() - ctx.started_at
        return elapsed_time > self.max_delay


class StopNever(BaseCondition):
    def __call__(self, ctx: RetryContext) -> bool:
        return False


never: Condition = StopNever()
after_attempt = StopAfterAttempt
after = StopAfter



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.
from abc import abstractmethod
from typing import Any, Optional, TypeVar

from typing_extensions import Self

from meatie import INF, Cache, Limiter, Rate, Request


class BaseClient:
    shared_cache: Cache

    def __init__(
        self,
        local_cache: Optional[Cache] = None,
        limiter: Optional[Limiter] = None,
    ):
        self.local_cache = local_cache if local_cache is not None else Cache()
        self.limiter = limiter if limiter is not None else Limiter(Rate.max, INF)

    def __init_subclass__(cls, **kwargs: Any) -> None:
        cls.shared_cache = Cache()

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self, exc_type: Optional[type[BaseException]], exc_val: Optional[BaseException], exc_tb: Any
    ) -> None:
        return None

    def authenticate(self, request: Request) -> None:
        pass

    @abstractmethod
    def send(self, request: Request) -> Any:
        ...


ClientType = TypeVar("ClientType", bound=BaseClient)



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

from typing import (
    Any,
    Protocol,
)

from meatie import AsyncResponse, Response
from meatie.internal.types import T


class TypeAdapter(Protocol[T]):
    def from_response(self, response: Response) -> T:
        ...

    async def from_async_response(self, response: AsyncResponse) -> T:
        ...

    def to_content(self, value: T) -> Any:
        ...



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.
from typing import NoReturn

from meatie import AsyncResponse, Response


class _ClientResponseAdapter:
    @staticmethod
    def from_response(response: Response) -> Response:
        return response

    @staticmethod
    def from_async_response(response: AsyncResponse) -> AsyncResponse:
        return response

    @staticmethod
    def to_content(value: Response) -> NoReturn:
        raise RuntimeError("JSON serialization is not supported")


ClientResponseAdapter = _ClientResponseAdapter()



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.
from typing import Any, Callable, Optional

import requests
from meatie.error import ParseResponseError, ResponseError


class Response:
    def __init__(
        self,
        response: requests.Response,
        get_json: Optional[Callable[[requests.Response], dict[str, Any]]] = None,
        get_text: Optional[Callable[[requests.Response], str]] = None,
    ) -> None:
        self.response = response
        if get_json is not None:
            self.get_json = get_json  # type: ignore[assignment]
        if get_text is not None:
            self.get_text = get_text  # type: ignore[assignment]

    @property
    def status(self) -> int:
        return self.response.status_code

    def read(self) -> bytes:
        try:
            return self.response.content
        except Exception as exc:
            raise ResponseError(self, exc) from exc

    def text(self) -> str:
        try:
            return self.get_text(self.response)
        except Exception as exc:
            raise ResponseError(self, exc) from exc

    def json(self) -> dict[str, Any]:
        try:
            return self.get_json(self.response)
        except requests.JSONDecodeError as exc:
            text = self.text()
            raise ParseResponseError(text, self, exc) from exc
        except Exception as exc:
            raise ResponseError(self, exc) from exc

    @classmethod
    def get_json(cls, response: requests.Response) -> dict[str, Any]:
        return response.json()

    @classmethod
    def get_text(cls, response: requests.Response) -> str:
        return response.text



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.
from json.decoder import JSONDecodeError
from typing import Any, Awaitable, Callable, Optional

from aiohttp import ClientResponse, ContentTypeError
from meatie.error import ParseResponseError, ResponseError


class Response:
    def __init__(
        self,
        response: ClientResponse,
        get_json: Optional[Callable[[ClientResponse], Awaitable[dict[str, Any]]]] = None,
        get_text: Optional[Callable[[ClientResponse], Awaitable[str]]] = None,
    ) -> None:
        self.response = response
        if get_json is not None:
            self.get_json = get_json  # type: ignore[assignment]
        if get_text is not None:
            self.get_text = get_text  # type: ignore[assignment]

    @property
    def status(self) -> int:
        return self.response.status

    async def read(self) -> bytes:
        try:
            return await self.response.content.read()
        except Exception as exc:
            raise ResponseError(self, exc) from exc

    async def text(self) -> str:
        try:
            return await self.get_text(self.response)
        except Exception as exc:
            raise ResponseError(self, exc) from exc

    async def json(self) -> dict[str, Any]:
        try:
            return await self.get_json(self.response)
        except JSONDecodeError as exc:
            text = await self.text()
            raise ParseResponseError(text, self, exc) from exc
        except ContentTypeError as exc:
            text = await self.text()
            raise ParseResponseError(text, self, exc) from exc

    @classmethod
    def get_json(cls, response: ClientResponse) -> Awaitable[dict[str, Any]]:
        return response.json()

    @classmethod
    def get_text(cls, response: ClientResponse) -> Awaitable[str]:
        return response.text(errors="replace")



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.
import json
from dataclasses import is_dataclass
from inspect import isclass
from typing import Any, Generic

import pydantic
import pydantic.json
from meatie import AsyncResponse, ParseResponseError, Response
from meatie.internal.types import T
from typing_extensions import is_typeddict

from . import JsonAdapter, TypeAdapter


class PydanticV1TypeAdapter(Generic[T]):
    def __init__(self, model_type: type[T]) -> None:
        self.model_type = model_type

    def from_response(self, response: Response) -> T:
        json_model = JsonAdapter.from_response(response)
        try:
            return pydantic.parse_obj_as(self.model_type, json_model)
        except pydantic.ValidationError as exc:
            text = response.text()
            raise ParseResponseError(text, response, exc) from exc

    async def from_async_response(self, response: AsyncResponse) -> T:
        json_model = await JsonAdapter.from_async_response(response)
        try:
            return pydantic.parse_obj_as(self.model_type, json_model)
        except pydantic.ValidationError as exc:
            text = await response.text()
            raise ParseResponseError(text, response, exc) from exc

    @staticmethod
    def to_content(value: T) -> Any:
        json_string = json.dumps(value, default=pydantic.json.pydantic_encoder)
        return json.loads(json_string)


class PydanticV1TypeAdapterFactory:
    @staticmethod
    def __call__(model_cls: type[T]) -> TypeAdapter[T]:
        return PydanticV1TypeAdapter(model_cls)

    @staticmethod
    def is_model_type(value: type[Any]) -> bool:
        return isclass(value) and (
            issubclass(value, pydantic.BaseModel) or is_dataclass(value) or is_typeddict(value)
        )



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

import time as sys_time
from dataclasses import dataclass
from typing import Any

from meatie import Duration, Time

__all__ = ["Cache"]


# TODO: implement limit on storage size to avoid memory leaks using OrderedDict
class Cache:
    def __init__(self) -> None:
        self.__storage: dict[str, _Record] = {}

    def load(self, key: str) -> Any:
        current_time = sys_time.monotonic()
        return self.load_at(current_time, key)

    def load_at(self, time: Time, key: str) -> Any:
        record_opt = self.__storage.get(key)
        if record_opt is None:
            return None

        if record_opt.expires_at < time:
            del self.__storage[key]
            return None

        record_opt.last_accessed_at = time
        return record_opt.value

    def store(self, key: str, value: Any, ttl: Duration) -> None:
        current_time = sys_time.monotonic()
        self.store_at(current_time, key, value, ttl)

    def store_at(self, time: Time, key: str, value: Any, ttl: Duration) -> Any:
        self.__storage[key] = _Record(value=value, expires_at=time + ttl, last_accessed_at=time)


@dataclass()
class _Record:
    value: Any
    expires_at: Time
    last_accessed_at: Time



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.
from typing import Union

from meatie import Context, EndpointDescriptor
from meatie.aio import AsyncContext, AsyncEndpointDescriptor
from meatie.internal.types import PT, T

__all__ = ["private"]


class PrivateOption:
    __PRIORITY = 50

    def __call__(
        self, descriptor: Union[EndpointDescriptor[PT, T], AsyncEndpointDescriptor[PT, T]]
    ) -> None:
        if isinstance(descriptor, EndpointDescriptor):
            return self.__sync_descriptor(descriptor)
        return self.__async_descriptor(descriptor)

    def __sync_descriptor(self, descriptor: EndpointDescriptor[PT, T]) -> None:
        descriptor.register_operator(PrivateOption.__PRIORITY, operator)

    def __async_descriptor(self, descriptor: AsyncEndpointDescriptor[PT, T]) -> None:
        descriptor.register_operator(PrivateOption.__PRIORITY, async_operator)


def operator(ctx: Context[T]) -> T:
    ctx.client.authenticate(ctx.request)
    return ctx.proceed()


async def async_operator(ctx: AsyncContext[T]) -> T:
    await ctx.client.authenticate(ctx.request)
    return await ctx.proceed()


private = PrivateOption()



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.
import asyncio
import time
from http import HTTPStatus
from typing import Awaitable, Callable, Optional, Union

from meatie import (
    Condition,
    Context,
    Duration,
    EndpointDescriptor,
    RetryContext,
    RetryError,
    has_status,
    never,
    zero,
)
from meatie.aio import AsyncContext, AsyncEndpointDescriptor
from meatie.internal.retry import WaitFunc
from meatie.internal.types import PT, T

__all__ = ["retry"]


class RetryOption:
    __PRIORITY = 75

    def __init__(
        self,
        on: Condition = has_status(HTTPStatus.TOO_MANY_REQUESTS),
        wait: WaitFunc = zero,
        stop: Condition = never,
        sleep_func: Union[
            Callable[[Duration], None], Callable[[Duration], Awaitable[None]], None
        ] = None,
    ) -> None:
        self.__on = on
        self.__wait = wait
        self.__stop = stop
        self.__sleep_func = sleep_func

    def __call__(
        self, descriptor: Union[EndpointDescriptor[PT, T], AsyncEndpointDescriptor[PT, T]]
    ) -> None:
        if isinstance(descriptor, EndpointDescriptor):
            return self.__sync_descriptor(descriptor)
        return self.__async_descriptor(descriptor)

    def __sync_descriptor(self, descriptor: EndpointDescriptor[PT, T]) -> None:
        sleep_func: Union[Callable[[float], None], None] = self.__sleep_func  # type: ignore[assignment]
        if sleep_func is None:
            sleep_func = time.sleep

        operator = RetryOperator(self.__on, self.__wait, self.__stop, sleep_func)
        descriptor.register_operator(RetryOption.__PRIORITY, operator)

    def __async_descriptor(self, descriptor: AsyncEndpointDescriptor[PT, T]) -> None:
        sleep_func: Union[Callable[[float], Awaitable[None]], None] = self.__sleep_func  # type: ignore[assignment]
        if sleep_func is None:
            sleep_func = asyncio.sleep

        operator = AsyncRetryOperator(self.__on, self.__wait, self.__stop, sleep_func)
        descriptor.register_operator(RetryOption.__PRIORITY, operator)


retry = RetryOption


class RetryOperator:
    def __init__(
        self,
        on: Condition,
        wait: WaitFunc,
        stop: Condition,
        sleep_func: Callable[[Duration], None],
    ) -> None:
        self.__condition = on
        self.__wait = wait
        self.__stop = stop
        self.__sleep_func = sleep_func

    def __call__(self, operation_ctx: Context[T]) -> T:
        retry_ctx = RetryContext(attempt_number=1, started_at=time.monotonic())
        last_result: Optional[T] = None
        stopped = False
        while not stopped:
            retry_ctx.error = None
            retry_ctx.response = None
            if retry_ctx.attempt_number > 1:
                wait_time = self.__wait(retry_ctx)
                if wait_time > 0.0:
                    self.__sleep_func(wait_time)

            try:
                last_result = operation_ctx.proceed()
                retry_ctx.response = operation_ctx.response
            except BaseException as exc:
                retry_ctx.error = exc

            if not self.__condition(retry_ctx):
                break

            retry_ctx.attempt_number += 1
            stopped = self.__stop(retry_ctx)

        if not stopped and last_result is not None:
            return last_result

        if retry_ctx.error is not None:
            raise retry_ctx.error
        raise RetryError()


class AsyncRetryOperator:
    def __init__(
        self,
        on: Condition,
        wait: WaitFunc,
        stop: Condition,
        sleep_func: Callable[[Duration], Awaitable[None]],
    ) -> None:
        self.__condition = on
        self.__wait = wait
        self.__stop = stop
        self.__sleep_func = sleep_func

    async def __call__(self, operation_ctx: AsyncContext[T]) -> T:
        retry_ctx = RetryContext(attempt_number=1, started_at=time.monotonic())
        last_result: Optional[T] = None
        stopped = False
        while not stopped:
            retry_ctx.error = None
            retry_ctx.response = None
            if retry_ctx.attempt_number > 1:
                wait_time = self.__wait(retry_ctx)
                if wait_time > 0.0:
                    await self.__sleep_func(wait_time)

            try:
                last_result = await operation_ctx.proceed()
                retry_ctx.response = operation_ctx.response
            except BaseException as exc:
                retry_ctx.error = exc

            if not self.__condition(retry_ctx):
                break

            retry_ctx.attempt_number += 1
            stopped = self.__stop(retry_ctx)

        if not stopped and last_result is not None:
            return last_result

        if retry_ctx.error is not None:
            raise retry_ctx.error
        raise RetryError()



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

import abc
import urllib.parse
from typing import Generic, Union

from meatie import Cache, Context, Duration, EndpointDescriptor, Request
from meatie.aio import AsyncContext, AsyncEndpointDescriptor
from meatie.internal.types import PT, T

__all__ = ["cache"]


class CacheOption:
    __PRIORITY = 25

    def __init__(self, ttl: Duration, shared: bool = False) -> None:
        self.ttl = ttl
        self.shared = shared

    def __call__(
        self, descriptor: Union[EndpointDescriptor[PT, T], AsyncEndpointDescriptor[PT, T]]
    ) -> None:
        if isinstance(descriptor, EndpointDescriptor):
            return self.__sync_descriptor(descriptor)
        return self.__async_descriptor(descriptor)

    def __sync_descriptor(self, descriptor: EndpointDescriptor[PT, T]) -> None:
        operator: Operator[T]
        if self.shared:
            operator = SharedOperator[T](self.ttl)
        else:
            operator = LocalOperator[T](self.ttl)
        descriptor.register_operator(CacheOption.__PRIORITY, operator)

    def __async_descriptor(self, descriptor: AsyncEndpointDescriptor[PT, T]) -> None:
        operator: AsyncOperator[T]
        if self.shared:
            operator = SharedAsyncOperator[T](self.ttl)
        else:
            operator = LocalAsyncOperator[T](self.ttl)
        descriptor.register_operator(CacheOption.__PRIORITY, operator)


cache = CacheOption


def get_key(request: Request) -> str:
    key = request.path
    if request.params:
        key += "?" + urllib.parse.urlencode(request.params)
    return key


class _Operator(Generic[T]):
    def __init__(self, ttl: Duration) -> None:
        self.ttl = ttl

    @abc.abstractmethod
    def _storage(self, ctx: Context[T]) -> Cache:
        ...


class Operator(Generic[T]):
    def __init__(self, ttl: Duration) -> None:
        self.ttl = ttl

    def __call__(self, ctx: Context[T]) -> T:
        storage = self._storage(ctx)
        key = get_key(ctx.request)
        value_opt = storage.load(key)
        if value_opt is not None:
            return value_opt

        value = ctx.proceed()
        storage.store(key, value, self.ttl)
        return value

    @abc.abstractmethod
    def _storage(self, ctx: Context[T]) -> Cache:
        ...


class LocalOperator(Operator[T]):
    def _storage(self, ctx: Context[T]) -> Cache:
        return ctx.client.local_cache


class SharedOperator(Operator[T]):
    def _storage(self, ctx: Context[T]) -> Cache:
        return ctx.client.shared_cache


class AsyncOperator(Generic[T]):
    def __init__(self, ttl: Duration) -> None:
        self.ttl = ttl

    async def __call__(self, ctx: AsyncContext[T]) -> T:
        storage = self._storage(ctx)
        key = get_key(ctx.request)
        value_opt = storage.load(key)
        if value_opt is not None:
            return value_opt

        value = await ctx.proceed()
        storage.store(key, value, self.ttl)
        return value

    @abc.abstractmethod
    def _storage(self, ctx: AsyncContext[T]) -> Cache:
        ...


class LocalAsyncOperator(AsyncOperator[T]):
    def _storage(self, ctx: AsyncContext[T]) -> Cache:
        return ctx.client.local_cache


class SharedAsyncOperator(AsyncOperator[T]):
    def _storage(self, ctx: AsyncContext[T]) -> Cache:
        return ctx.client.shared_cache



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.
from typing import Any, TypeVar

from meatie import AsyncResponse, Request, Response
from typing_extensions import Optional, ParamSpec, Protocol, Self

__all__ = [
    "VT",
    "T",
    "T_In",
    "T_Out",
    "PT",
    "RequestBodyType",
    "ResponseBodyType",
    "Client",
    "ClientType",
    "Context",
    "AsyncClient",
    "AsyncClientType",
    "AsyncContext",
]

VT = TypeVar("VT")
T = TypeVar("T")
T_In = TypeVar("T_In", contravariant=True)
T_Out = TypeVar("T_Out", covariant=True)
PT = ParamSpec("PT")
RequestBodyType = TypeVar("RequestBodyType", contravariant=True)
ResponseBodyType = TypeVar("ResponseBodyType", covariant=True)


class Client(Protocol):
    def send(self, request: Request) -> Response:
        ...

    def __enter__(self) -> Self:
        ...

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Any,
    ) -> None:
        ...


ClientType = TypeVar("ClientType", bound=Client, covariant=True)


class Context(Protocol[ClientType, ResponseBodyType]):
    @property
    def client(self) -> ClientType:
        ...

    @property
    def request(self) -> Request:
        ...

    @property
    def response(self) -> Optional[Response]:
        ...

    def proceed(self) -> ResponseBodyType:
        ...


class AsyncClient(Protocol):
    async def send(self, request: Request) -> AsyncResponse:
        ...

    async def __aenter__(self) -> Self:
        ...

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Any,
    ) -> None:
        ...


AsyncClientType = TypeVar("AsyncClientType", bound=AsyncClient, covariant=True)


class AsyncContext(Protocol[AsyncClientType, ResponseBodyType]):
    @property
    def client(self) -> AsyncClientType:
        ...

    @property
    def request(self) -> Request:
        ...

    @property
    def response(self) -> Optional[AsyncResponse]:
        ...

    async def proceed(self) -> ResponseBodyType:
        ...



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.
from dataclasses import dataclass
from typing import Any, Optional, Protocol, Union, runtime_checkable

from typing_extensions import Literal

Method = Literal["GET", "HEAD", "POST", "PUT", "DELETE", "OPTIONS", "TRACE", "PATCH"]

Duration = float
Time = float

MINUTE = 60
HOUR = 3600
DAY = 86400

INF = float("inf")


@dataclass()  # TODO: change to an ordinary class
class Request:
    method: Method
    path: str
    params: dict[str, Union[str, int]]
    headers: dict[str, Union[str, bytes]]
    data: Optional[Union[str, bytes]] = None
    json: Any = None


@runtime_checkable
class AsyncResponse(Protocol):
    @property
    def status(self) -> int:
        ...

    async def read(self) -> bytes:
        ...

    async def text(self) -> str:
        ...

    async def json(self) -> dict[str, Any]:
        ...


@runtime_checkable
class Response(Protocol):
    @property
    def status(self) -> int:
        ...

    def read(self) -> bytes:
        ...

    def text(self) -> str:
        ...

    def json(self) -> dict[str, Any]:
        ...



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

# isort: skip_file
from .context import RetryContext
from .condition import AndCondition, BaseCondition, Condition, OrCondition
from .retry import always, has_status, has_exception_type, has_exception_cause_type
from .stop import after, after_attempt, never
from .wait import zero, uniform, exponential, fixed, jit, WaitFunc

__all__ = [
    "RetryContext",
    "AndCondition",
    "BaseCondition",
    "Condition",
    "OrCondition",
    "after",
    "after_attempt",
    "always",
    "never",
    "has_status",
    "has_exception_type",
    "has_exception_cause_type",
    "WaitFunc",
    "zero",
    "uniform",
    "exponential",
    "fixed",
    "jit",
]



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.
from typing import Any, Optional, Union

from meatie import AsyncResponse, Response


class MeatieError(Exception):
    ...


class RetryError(MeatieError):
    ...


class RateLimitExceeded(MeatieError):
    ...


class TransportError(MeatieError):
    ...


class ProxyError(TransportError):
    ...


class ServerError(TransportError):
    ...


class Timeout(ServerError):
    ...


class RequestError(MeatieError):
    ...


class ResponseError(MeatieError):
    def __init__(self, response: Union[Response, AsyncResponse], *args: Any) -> None:
        super().__init__(*args)

        self.response = response


class HttpStatusError(ResponseError):
    ...


class ParseResponseError(ResponseError):
    def __init__(
        self,
        text: str,
        response: Any,
        cause: Optional[BaseException] = None,
    ) -> None:
        super().__init__(response, cause)

        self.text = text



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.
from typing import Any, Awaitable, Callable, Optional, Union

from meatie import EndpointDescriptor
from meatie.aio import AsyncEndpointDescriptor
from meatie.internal.types import PT, T

__all__ = ["body"]


class BodyOption:
    __slots__ = ("json", "text")

    def __init__(
        self,
        json: Optional[
            Union[Callable[[Any], dict[str, Any]], Callable[[Any], Awaitable[dict[str, Any]]]]
        ] = None,
        text: Optional[Union[Callable[[Any], str], Callable[[Any], Awaitable[str]]]] = None,
    ) -> None:
        """
        Customize handling of HTTP response body.

        :param json: function to apply on the HTTP response to extract json, otherwise use the default method
         implemented in the client library
        :param text: function to apply on the HTTP response to extract response text, otherwise use the default method
         implemented in the client library
        """

        self.json = json
        self.text = text

    def __call__(
        self, descriptor: Union[EndpointDescriptor[PT, T], AsyncEndpointDescriptor[PT, T]]
    ) -> None:
        descriptor.get_text = self.text
        descriptor.get_json = self.json


body = BodyOption



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

import re
from typing import (
    Any,
)

from typing_extensions import Self

_param_pattern = re.compile(r"{(?P<name>[a-zA-Z_]\w*?)}")


class PathTemplate:
    __slots__ = ("template", "parameters")

    def __init__(self, template: str, parameters: list[str]) -> None:
        self.template = template
        self.parameters = parameters

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, PathTemplate):
            return self.template == other.template and self.parameters == other.parameters
        if isinstance(other, str):
            return self.template == other
        return False

    def __hash__(self) -> int:
        return hash(self.template)

    def __contains__(self, item: str) -> bool:
        return item in self.parameters

    def __str__(self) -> str:
        return self.template

    def format(self, **kwargs: dict[str, Any]) -> str:
        return self.template.format(**kwargs)

    @classmethod
    def from_string(cls, template: str) -> Self:
        parameters = [match.group("name") for match in _param_pattern.finditer(template)]
        return cls(template, parameters)



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

__all__ = ["cache", "limit", "private", "retry", "body"]

from .body_option import body
from .cache_option import cache
from .limit_option import limit
from .private_option import private
from .retry_option import retry



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

import re
from typing import get_args

from meatie import Method

_method_pattern_pairs = [
    (method, re.compile("^" + method, re.IGNORECASE)) for method in get_args(Method)
]


def get_method(name: str, default: Method = "GET") -> Method:
    for method, pattern in _method_pattern_pairs:
        if re.match(pattern, name):
            return method
    return default



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.
import abc
import random
from abc import abstractmethod
from typing import Callable

from meatie import HOUR, Duration
from typing_extensions import Union

from . import RetryContext

__all__ = ["zero", "uniform", "jit", "exponential", "fixed", "WaitFunc"]

WaitFunc = Callable[[RetryContext], Duration]


class BaseWait(abc.ABC):
    @abstractmethod
    def __call__(self, ctx: RetryContext) -> Duration:
        raise NotImplementedError()

    def __add__(self, other: WaitFunc) -> "BaseWait":
        return WaitSum(self, other)


class WaitSum(BaseWait):
    def __init__(self, *waits: WaitFunc) -> None:
        self.waits = waits

    def __call__(self, ctx: RetryContext) -> Duration:
        return sum(wait(ctx) for wait in self.waits)


class WaitZero(BaseWait):
    def __call__(self, ctx: RetryContext) -> Duration:
        return 0.0


class WaitExponential:
    def __init__(
        self, exp_base: float = 2.0, multiplier: float = 2.0, lb: float = 0.0, ub: float = HOUR
    ) -> None:
        self.exp_base = exp_base
        self.multiplier = multiplier
        self.lb = max(lb, 0.0)
        self.ub = ub

    def __call__(self, ctx: RetryContext) -> Duration:
        try:
            result = self.multiplier * self.exp_base ** (ctx.attempt_number - 1)
        except OverflowError:
            return self.ub
        return max(self.lb, min(result, self.ub))


class WaitFixed:
    def __init__(self, delays: Union[Duration, tuple[Duration, ...]]) -> None:
        if isinstance(delays, Duration):
            self.__delays: tuple[Duration, ...] = (delays,)
        else:
            self.__delays = delays

    def __call__(self, ctx: RetryContext) -> Duration:
        step = min(ctx.attempt_number, len(self.__delays)) - 1
        if step >= 0:
            return self.__delays[step]
        return 0.0


class WaitUniform:
    def __init__(self, lb: Duration, ub: Duration) -> None:
        self.lb = lb
        self.ub = ub

    def __call__(self, ctx: RetryContext) -> Duration:
        return random.uniform(self.lb, self.ub)


class WaitJit(WaitUniform):
    def __init__(self, ub: Duration) -> None:
        super().__init__(0.0, ub)


zero: WaitFunc = WaitZero()
uniform = WaitUniform
jit = WaitJit
exponential = WaitExponential
fixed = WaitFixed



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.
from meatie import AsyncResponse, Response
from meatie.error import ResponseError


class _StringAdapter:
    @staticmethod
    def from_response(response: Response) -> str:
        try:
            return response.text()
        except ValueError as exc:
            raise ResponseError(response, exc) from exc

    @staticmethod
    async def from_async_response(response: AsyncResponse) -> str:
        try:
            return await response.text()
        except ValueError as exc:
            raise ResponseError(response, exc) from exc

    @staticmethod
    def to_content(value: str) -> str:  # pragma: no cover
        return value


StringAdapter = _StringAdapter()



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.


import time as sys_time
from typing import Optional

from meatie import Time

from . import Rate, Reservation, Tokens


class Limiter:
    __slots__ = ("rate", "capacity", "__last_tokens", "__last_time")

    def __init__(
        self,
        rate: Rate,
        capacity: Tokens,
        init_tokens: Optional[Tokens] = None,
        init_time: Optional[Time] = None,
    ) -> None:
        self.rate = rate
        self.capacity = capacity
        self.__last_tokens = init_tokens if init_tokens is not None else capacity
        self.__last_time = init_time if init_time is not None else sys_time.monotonic()

    def reserve_now(self, tokens: Tokens) -> Reservation:
        return self.reserve_at(sys_time.monotonic(), tokens)

    def reserve_at(self, time: Time, tokens: Tokens) -> Reservation:
        if tokens > self.capacity:
            raise ValueError(
                f"amount of requested tokens ({tokens}) exceed the limit ({self.capacity})"
            )

        available = self.__advance_until(time)
        remaining = available - tokens

        if remaining < 0:
            wait_duration = self.rate.duration_from_tokens(-remaining)
        else:
            wait_duration = 0.0

        result = Reservation(ready_at=time + wait_duration, tokens=tokens)
        self.__last_time = time
        self.__last_tokens = remaining
        return result

    def __advance_until(self, time: Time) -> Time:
        last_time = self.__last_time
        if time < last_time:
            last_time = time

        duration = time - last_time
        delta = self.rate.tokens_from_duration(duration)
        tokens = self.__last_tokens + delta
        if tokens > self.capacity:
            return self.capacity
        return tokens



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.
import inspect
from typing import Any, Callable, Optional, get_args

from typing_extensions import Self

__all__ = ["api_ref", "ApiReference"]


class ApiReference:
    __slots__ = ("name", "fmt")

    def __init__(
        self, name: Optional[str] = None, fmt: Optional[Callable[[Any], Any]] = None
    ) -> None:
        """
        Customize handling of a parameter in the HTTP request.

        :param name: name of the query parameter in the HTTP request, if not set the name of the Python parameter is
         used by default
        :param fmt: conversion function to apply on the parameter value before sending the HTTP request
        """

        self.name = name
        self.fmt = fmt

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other: Any) -> bool:
        # Python interpreter seems to reuse annotations, for that reason we also need to include the formatter
        if isinstance(other, ApiReference):
            return self.name == other.name and self.fmt == other.fmt
        return False

    @classmethod
    def from_signature(cls, parameter: inspect.Parameter) -> Self:
        for arg in get_args(parameter.annotation):
            if isinstance(arg, cls):
                if arg.name is None:
                    return cls(name=parameter.name, fmt=arg.fmt)
                return arg
        return cls(name=parameter.name, fmt=None)


api_ref = ApiReference



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.
import asyncio
from typing import Any, Optional

import aiohttp
from meatie import (
    AsyncResponse,
    BaseAsyncClient,
    Cache,
    MeatieError,
    ProxyError,
    Request,
    ServerError,
    Timeout,
    TransportError,
)

from . import Response


class Client(BaseAsyncClient):
    def __init__(
        self,
        session: aiohttp.ClientSession,
        session_params: Optional[dict[str, Any]] = None,
        local_cache: Optional[Cache] = None,
        limiter: Optional[Any] = None,
    ) -> None:
        super().__init__(local_cache, limiter)

        self.session = session
        self.session_params = session_params if session_params else {}

    async def send(self, request: Request) -> AsyncResponse:
        kwargs: dict[str, Any] = self.session_params.copy()

        if request.data is not None:
            kwargs["data"] = request.data

        if request.json is not None:
            kwargs["json"] = request.json

        if request.headers:
            kwargs["headers"] = request.headers

        if request.params:
            kwargs["params"] = request.params

        try:
            response = await self.session.request(request.method, request.path, **kwargs)
        except (aiohttp.ClientProxyConnectionError, aiohttp.ClientHttpProxyError) as exc:
            raise ProxyError(exc) from exc
        except (
            aiohttp.ServerConnectionError,
            aiohttp.ServerDisconnectedError,
            aiohttp.ClientConnectorError,
            aiohttp.ClientConnectionError,
            aiohttp.ClientConnectorSSLError,
            aiohttp.ClientConnectorCertificateError,
        ) as exc:
            raise ServerError(exc) from exc
        except (aiohttp.ServerTimeoutError, asyncio.TimeoutError) as exc:
            raise Timeout(exc) from exc
        except (aiohttp.TooManyRedirects, aiohttp.ClientError) as exc:
            raise TransportError(exc) from exc
        except Exception as exc:
            raise MeatieError(exc) from exc
        return Response(response)

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Any,
    ) -> None:
        await self.close()

    async def close(self) -> None:
        await self.session.close()



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.
from json import JSONDecodeError
from typing import Any

from aiohttp import ClientError, ContentTypeError
from meatie import AsyncResponse, ParseResponseError, Response, ResponseError


class _JsonAdapter:
    @staticmethod
    def from_response(response: Response) -> Any:
        try:
            return response.json()
        except ContentTypeError as exc:
            raise ResponseError(response, exc) from exc
        except JSONDecodeError as exc:
            text = response.text()
            raise ParseResponseError(text, response, exc) from exc

    @staticmethod
    async def from_async_response(response: AsyncResponse) -> Any:
        try:
            return await response.json()
        except ContentTypeError as exc:
            raise ResponseError(response, exc) from exc
        except JSONDecodeError as exc:
            text = await response.text()
            raise ParseResponseError(text, response, exc) from exc
        except ClientError as exc:
            raise ResponseError(response, exc) from exc

    @staticmethod
    def to_content(value: Any) -> Any:
        return value


JsonAdapter = _JsonAdapter()



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.
import inspect
from collections.abc import Callable
from typing import (
    Any,
    Optional,
    Union,
    get_type_hints,
)

from meatie import AsyncEndpointDescriptor, EndpointDescriptor, Method
from meatie.internal.adapter import TypeAdapter, get_adapter
from meatie.internal.template import PathTemplate, RequestTemplate
from meatie.internal.types import PT, T


def endpoint(
    path: str,
    *args: Any,
    method: Optional[Method] = None,
) -> Callable[[Callable[..., T]], Callable[..., T]]:
    def class_descriptor(func: Callable[PT, T]) -> Callable[PT, T]:
        path_template = PathTemplate.from_string(path)

        signature = inspect.signature(func)
        type_hints = get_type_hints(func)
        request_template: RequestTemplate[T] = RequestTemplate.from_signature(
            signature, type_hints, path_template, method
        )

        return_type = type_hints["return"]
        response_decoder: TypeAdapter[T] = get_adapter(return_type)

        descriptor: Union[EndpointDescriptor[PT, T], AsyncEndpointDescriptor[PT, T]]
        is_coroutine = inspect.iscoroutinefunction(func)
        if is_coroutine:
            descriptor = AsyncEndpointDescriptor[PT, T](request_template, response_decoder)
        else:
            descriptor = EndpointDescriptor[PT, T](request_template, response_decoder)

        for option in args:
            option(descriptor)

        return descriptor  # type: ignore[return-value]

    return class_descriptor



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.


from dataclasses import dataclass

from meatie import Time

Tokens = float


@dataclass()
class Reservation:
    tokens: Tokens
    ready_at: Time
